#include <Arduino.h>
#include "globals.h"
#include "swerve.h"
#include <math.h>

template <typename T> int signum(T val) {
    return (T(0) < val) - (val < T(0));
}

// 3-mode drive control system based on https://compendium.readthedocs.io/en/latest/tasks/drivetrains/swerve.html

float closestAngle(float current, float target)
{
        while(current > 360) {
            current -= 360.0;
        }
        current = 360 - current;
        while(target > 360) {
            target -= 360.0;
        }
        target = 360 - target;
        // get direction
        float dir = target - current;

        // convert from -360 to 360 to -180 to 180
        if (abs(dir) > 180.0)
        {
                dir = -(signum(dir) * 360.0) + dir;
        }
        return dir;
}

// TODO as of 20230926
// This function calculates the robot's current speed and attempts to modify the current state of the drive towards the target drive state
swerve_drive updateSwerveCommand(swerve_drive input) {
    swerve_drive out = input;

    // Monitor encoder values for each wheel and place them in the encoder_buffer arrays in the swerve_drive struct
        // TODO, get encoder values for each drive motor and place them in the arrays
        // TODO, get time in milliseconds (either unix millis or relative millis to pico startup)

    // Calculate the robot's average speed over the last few encoder samples
        // TODO
    
    // Determine how much it is safe to change the orientation of the wheels, given the robot's current speed
        // TODO

    // Update every "current" variable in the swerve_drive struct with new data
        // TODO
    
    // Set the new commanded orientation of the rotation motors and the new commanded drive motor speed
        // TODO

    return out;
}

/*
// TODO as of 20230923 for setDirection() : fix to work with modifications made to swerve_drive struct on 20230922
swerve_drive setDirection(swerve_drive input, float setpoint)
{
    swerve_drive out = input;
    float currentAngle = input.current_spin_location;
    // find closest angle to setpoint
    float setpointAngle = closestAngle(currentAngle, setpoint);
    // find closest angle to setpoint + 180
    float setpointAngleFlipped = closestAngle(currentAngle, setpoint + 180.0);
    // if the closest angle to setpoint is shorter
    if (abs(setpointAngle) <= abs(setpointAngleFlipped))
    {
        // unflip the motor direction use the setpoint
        out.target_spin_direction = CLOCKWISE;
        out.target_spin_location = (currentAngle + setpointAngle);
    }
    // if the closest angle to setpoint + 180 is shorter
    else
    {
        // flip the motor direction and use the setpoint + 180
        out.target_spin_direction = ANTICLOCKWISE;
        out.target_spin_location = (currentAngle + setpointAngleFlipped);
    }
    return out;    
}
*/

swerve_drive tempTranslationDrive(swerve_drive input, float target_speed, float target_angle) // Temporary implementation for translation drive mode
{
    swerve_drive out = input;

    float normalized_target_angle = normalizeAngle(target_angle); // Normalize the target angle

    out = setTargetSpin(out, normalized_target_angle, normalized_target_angle, normalized_target_angle, normalized_target_angle); // Set the target angle for each rotation motor
    out = setMotorCoefficients(out, 1.0, 1.0, 1.0, 1.0); // Set the motor speed coefficients to 1 for all motors
    out = setDriveTargetPower(out, target_speed); // Set the power
    return out;
}

/*
    TODO for tempRotationDrive() as of 20230923: add functionality to track the robot body's rotation relative to the field through an IMU or encoders somehow,
    and allow the option to modify the spin offset to compensate for it.
*/
swerve_drive tempRotationDrive(swerve_drive input, float target_speed) // Temporary implementation for rotation drive mode (rotating in place), positive speed is clockwise, negative speed is counterclockwise
{
    swerve_drive out = input;

    //float normalized_target_angle = normalizeAngle(target_angle); // Normalize the target angle

    out = setTargetSpin(out, 45.0, 135.0, 225.0, 315.0); // Set the target angle for each rotation motor
    out = setMotorCoefficients(out, 1.0, 1.0, 1.0, 1.0); // Set the motor speed coefficients to 1 for all motors
    out = setDriveTargetPower(out, target_speed); // Set the power

    return out;
}

/*
    TODO for tempBasicDrive() as of 20230923: add functionality to track the robot body's rotation relative to the field through an IMU or encoders somehow,
    and allow the option to modify the spin offset to compensate for it.
*/
swerve_drive tempBasicDrive(swerve_drive input, float target_speed, float target_angle) // Temporary implementation for basic drive mode
{
    swerve_drive out = input;

    float normalized_target_angle = normalizeAngle(target_angle); // Normalize the target angle
    byte drive_condition = identifyBasicDriveCondition(target_speed, normalized_target_angle); // Identify the drive condition

    // Find inner_difference_angle
    // Measure from the target angle to the nearest section of the y-axis (like on a unit circle), this is for the inner wheels in the turn (wheels closer to the center of the turning circle)
    float inner_difference_angle;
    float semicircle_normalized_target_angle = fmod(normalized_target_angle, 180.0);
    if(semicircle_normalized_target_angle < 90) { // inner_difference_angle is less than 90
        inner_difference_angle = semicircle_normalized_target_angle;
    } else { // inner_difference_angle is greater than 90
        inner_difference_angle = 180 - semicircle_normalized_target_angle;
    }

    // Find the turning radius for the inner wheels, multiply by the distance between the robot's wheels to get the actual distance
    float inner_wheel_turning_radius = tan(RADIANS_PER_DEGREE * (90.0 - inner_difference_angle)) / 2.0;

    // Find the turning radius for the outer wheels, multiply by the distance between the robot's wheels to get the actual distance
    float outer_wheel_turning_radius = inner_wheel_turning_radius + 1.0;
    
    // Find the inner drive motor coefficient (to slow down the inner motors to compensate for a smaller turning radius), this is the ratio bewteen the inner and outer turning radii
    float inner_drive_coefficient = outer_wheel_turning_radius / inner_wheel_turning_radius;

    // Find outer_difference_angle
    // Measure from the target angle to the nearest y-axis, this is for the outer wheels in the turn (wheels further from the center of the turning circle)
    float outer_difference_angle = 90.0 - (DEGREES_PER_RADIAN * atan(2.0 * outer_wheel_turning_radius));

    // Calculate the target spin angle for the outer wheels in the turn from the outer_difference_angle
    float outer_target_angle;
    int normalized_target_quadrant = ((int) normalized_target_angle ) / 90;
    if (normalized_target_quadrant % 2 == 0) { // Quadrants 0 or 2
        outer_target_angle = outer_difference_angle + (float) (90 * normalized_target_quadrant);
    } else { // Quadrants 1 or 3
        outer_target_angle = -outer_difference_angle + (float) (90 * (normalized_target_quadrant + 1));
    }

    // Prepare these variables to be written to in the switch statement below
    // Set these to 0.0, so the robot will stop if none of the cases are hit for some reason
    float front_left_coefficient = 0.0;
    float front_right_coefficient = 0.0;

    // Determine the speed coefficient for each motor (how fast motors are spinning relative to each other)
    switch(drive_condition) {
        case DRIVE_BASIC_STOP: // Motors are stopping, set target_speed to 0
            front_left_coefficient = 0.0;
            front_right_coefficient = 0.0;
            target_speed = 0.0;
            break;
        case DRIVE_BASIC_FORWARD: case DRIVE_BASIC_BACKWARD: // Motors are all spinning at the same speed since the robot is going in a straight line
            front_left_coefficient = 1.0;
            front_right_coefficient = 1.0;
            break;
        case DRIVE_BASIC_FRONTLEFT: case DRIVE_BASIC_BACKLEFT: // Left wheels are the inner wheels in the turn, so they will go slower
            front_left_coefficient = inner_drive_coefficient;
            front_right_coefficient = 1.0;
            break;
        case DRIVE_BASIC_FRONTRIGHT: case DRIVE_BASIC_BACKRIGHT: // Right wheels are the inner wheels in the turn, so they will go slower
            front_left_coefficient = 1.0;
            front_right_coefficient = inner_drive_coefficient;
            break;
    }
    // Set the coefficients for the remaining 2 motors
    float back_left_coefficient = front_left_coefficient; // Front and back coefficients are the same for each side (left and right)
    float back_right_coefficient = front_right_coefficient; // Front and back coefficients are the same for each side (left and right)

    // Set target wheel spin angles
    switch(drive_condition) {
        case DRIVE_BASIC_STOP: // Do not modify wheel target angles, robot is stopping
            break;
        case DRIVE_BASIC_FORWARD: case DRIVE_BASIC_BACKWARD: // All wheels facing in the same direction since the robot is going in a straight line
            out = setTargetSpin(out, normalized_target_angle, normalized_target_angle, normalized_target_angle, normalized_target_angle); 
            break;
        case DRIVE_BASIC_FRONTLEFT: case DRIVE_BASIC_BACKLEFT: // Vary the angles of the wheels since the robot is turning, left wheels are the inner wheels in the turn
            out = setTargetSpin(out, normalized_target_angle, outer_difference_angle, normalized_target_angle, outer_difference_angle);
            break;
        case DRIVE_BASIC_FRONTRIGHT: case DRIVE_BASIC_BACKRIGHT: // Vary the angles of the wheels since the robot is turning, right wheels are the inner wheels in the turn
            out = setTargetSpin(out, outer_difference_angle, normalized_target_angle, outer_difference_angle, normalized_target_angle);
            break;
    }

    // Wrap up, set motor speed and speed coefficient values
    out = setMotorCoefficients(out, front_left_coefficient, front_right_coefficient, back_left_coefficient, back_right_coefficient); // Set motor speed coefficients
    out = setDriveTargetPower(out, target_speed); // Set the power

    return out;
}

byte identifyBasicDriveCondition(float target_speed, float target_angle) // Identify the condition in which the basic drive mode will be operating
{
    if(target_speed <= 0.0 || target_speed >= 127.0) { // Speed is stopped, negative, or too high
        return DRIVE_BASIC_STOP;
    }
    target_angle = normalizeAngle(target_angle); // Normalize the target angle
    if(target_angle == 0) {
        return DRIVE_BASIC_FORWARD;
    }
    if(target_angle == 180) {
        return DRIVE_BASIC_BACKWARD;
    }
    int target_quadrant = ((int) target_angle) / 90; // Which quadrant is the target angle in (quadrant 0 is northeast, increases clockwise)
    byte drive_condition = DRIVE_BASIC_STOP; // Keep the drive stopped if the normalized target angle somehow exceeded 360.
    switch(target_quadrant) {
        case 0:
            drive_condition = DRIVE_BASIC_FRONTRIGHT;
            break;
        case 1:
            drive_condition = DRIVE_BASIC_BACKRIGHT;
            break;
        case 2:
            drive_condition = DRIVE_BASIC_BACKLEFT;
            break;
        case 3:
            drive_condition = DRIVE_BASIC_FRONTLEFT;
            break;
        default: // Unreachable, assuming that the target angle is less than 360
            drive_condition = DRIVE_BASIC_STOP;
            break;
    }
    return drive_condition;
}

float normalizeAngle(float angle) { // Takes an input angle and normalizes it to an angle between 0.0 and 360.0 degrees, results excluding exactly 360.0 degrees
    angle = fmod(angle, 360);
    if(angle < 0.0) {
        angle += 360;
    }
    return angle;
}

swerve_drive setMotorCoefficients(swerve_drive input, float front_left, float front_right, float back_left, float back_right) { // Set the motor speed coefficients for each motor
    swerve_drive out = input;
    out.front_left_coefficient = front_left;
    out.front_right_coefficient = front_right;
    out.back_left_coefficient = back_left;
    out.back_right_coefficient = back_right;
    return out;
}

swerve_drive setTargetSpin(swerve_drive input, float front_left, float front_right, float back_left, float back_right) { // Set the target spin for each wheel
    swerve_drive out = input;
    out.front_left_target_spin = front_left + input.spin_offset;
    out.front_right_target_spin = front_right + input.spin_offset;
    out.back_left_target_spin = back_left + input.spin_offset;
    out.back_right_target_spin = back_right + input.spin_offset;
    return out;
}

swerve_drive setSpinOffset(swerve_drive input, float new_spin_offset) { // Set a new spin offset, and maintain the current target spin on each motor relative to the robot body as the offset is changed
    swerve_drive out = input;
    
    float delta_spin_offset = new_spin_offset - input.spin_offset;
    
    out.front_left_target_spin = input.front_left_target_spin - delta_spin_offset;
    out.front_right_target_spin = input.front_left_target_spin - delta_spin_offset;
    out.back_left_target_spin = input.front_left_target_spin - delta_spin_offset;
    out.back_right_target_spin = input.front_left_target_spin - delta_spin_offset;
    return out;
}

swerve_drive setDriveTargetPower(swerve_drive input, float target_drive_power) { // Set a new drive power
    swerve_drive out = input;
    out.target_drive_power = target_drive_power;
    return out;
}



// Modulus Plus: Drive control planning: v1 on 20230922
/*

315   000   045
       
270    ^    090

225   180   135

If speed AND any wheel's target-current angle are BOTH above a certain threshold, decelerate before spinning the wheel

Variables to monitor and control:
    Drive modes:
        Translation (Linear motion in any direction while preserving body orientation)
        Basic (Forwards or backwards, may turn)
        Rotation (rotation in place)
    Rotation angle (in Translation or Basic drive mode)
    Drive speed
    Drive forward orientation relative to robot body

Common Controls:
    Button 1:
        Switch between drive modes
    Button 2:
        Toggle preservation of drive forward orientation relative to field
        When off, the drive forward direction is locked relative to the body
        When on, the drive forward direction is locked relative to the field
    D-Pad (While button 2 is pressed):
        Reset drive orientation relative to body
        Set direction to the corresponding d-pad button direction

Translation Mode Controls:
    Joystick A:
        Vector Amplitude:   Drive speed
        Vector Angle:       Drive angle

Basic Mode Controls:
    Joystick A:
        Vector Amplitude:   Drive speed
        Vector Angle:       Drive angle

Rotation Mode Controls:
    Joystick A:
        Left/Right:         Rotation speed

Rotation drive mode wheel orientation
Turning in place
//   \\

\\   //


Translation or Basic drive mode wheel orientation, with a rotation angle of 0
Driving in a straight line
||   ||


||   ||

Translation drive mode wheel orientation, with a rotation angle of 90 or 270
Driving to the left or right
_     _
-     -

_     _
-     -

Translation drive mode wheel orientation, with a rotation angle of 45
Driving forward and to the right
//   //

//   //

Basic drive mode wheel orientation with a rotation angle of 315
Driving forwards while turning (to the left in this example)
\\   \\

//   //

*/
